@@ -22,8 +22,9 @@ import typing as T
 from dataclasses import dataclass
 from enum import Enum
 
+from . import formatter
 from .gir import *
-from .utils import Colors
+from .utils import Colors, escape_quote
 from .xml_reader import Element, parse, parse_string
 
 __all__ = ["decompile"]
@@ -68,7 +69,9 @@ class DecompileCtx:
                 for ns, namespace in self.gir.namespaces.items()
             ]
         )
-        return imports + "\n" + self._result
+        full_string = imports + "\n" + self._result
+        formatted_string = formatter.format(full_string)
+        return formatted_string
 
     def type_by_cname(self, cname: str) -> T.Optional[GirType]:
         if type := self.gir.get_type_by_cname(cname):
@@ -96,35 +99,11 @@ class DecompileCtx:
         self._blocks_need_end[-1] = text
 
     def print(self, line: str, newline: bool = True) -> None:
-        if line == "}" or line == "]":
-            self._indent -= 1
-
-        # Add blank lines between different types of lines, for neatness
-        if newline:
-            if line == "}" or line == "]":
-                line_type = LineType.BLOCK_END
-            elif line.endswith("{") or line.endswith("]"):
-                line_type = LineType.BLOCK_START
-            elif line.endswith(";"):
-                line_type = LineType.STMT
-            else:
-                line_type = LineType.NONE
-            if (
-                line_type != self._last_line_type
-                and self._last_line_type != LineType.BLOCK_START
-                and line_type != LineType.BLOCK_END
-            ):
-                self._result += "\n"
-            self._last_line_type = line_type
-
-        self._result += ("  " * self._indent) + line
-        if newline:
-            self._result += "\n"
+        self._result += line
 
         if line.endswith("{") or line.endswith("["):
             if len(self._blocks_need_end):
                 self._blocks_need_end[-1] = _CLOSING[line[-1]]
-            self._indent += 1
 
     def print_attribute(self, name: str, value: str, type: GirType) -> None:
         def get_enum_name(value):
@@ -138,7 +117,7 @@ class DecompileCtx:
             return value.replace("-", "_")
 
         if type is None:
-            self.print(f'{name}: "{escape_quote(value)}";')
+            self.print(f"{name}: {escape_quote(value)};")
         elif type.assignable_to(FloatType()):
             self.print(f"{name}: {value};")
         elif type.assignable_to(BoolType()):
@@ -157,7 +136,7 @@ class DecompileCtx:
                 self.gir.namespaces["Gtk"].lookup_type("Gtk.ShortcutTrigger")
             )
         ):
-            self.print(f'{name}: "{escape_quote(value)}";')
+            self.print(f"{name}: {escape_quote(value)};")
         elif value == self.template_class:
             self.print(f"{name}: template;")
         elif type.assignable_to(
@@ -170,7 +149,7 @@ class DecompileCtx:
         elif isinstance(type, Enumeration):
             self.print(f"{name}: {get_enum_name(value)};")
         else:
-            self.print(f'{name}: "{escape_quote(value)}";')
+            self.print(f"{name}: {escape_quote(value)};")
 
 
 def _decompile_element(
@@ -253,15 +232,6 @@ def decompiler(tag, cdata=False):
     return decorator
 
 
-def escape_quote(string: str) -> str:
-    return (
-        string.replace("\\", "\\\\")
-        .replace("'", "\\'")
-        .replace('"', '\\"')
-        .replace("\n", "\\n")
-    )
-
-
 @decompiler("interface")
 def decompile_interface(ctx, gir):
     return gir
@@ -289,11 +259,11 @@ def decompile_translatable(
             comments = f"/* Translators: {comments} */"
 
         if context is not None:
-            return comments, f'C_("{escape_quote(context)}", "{escape_quote(string)}")'
+            return comments, f"C_({escape_quote(context)}, {escape_quote(string)})"
         else:
-            return comments, f'_("{escape_quote(string)}")'
+            return comments, f"_({escape_quote(string)})"
     else:
-        return comments, f'"{escape_quote(string)}"'
+        return comments, f"{escape_quote(string)}"
 
 
 @decompiler("property", cdata=True)
@@ -334,7 +304,7 @@ def decompile_property(
             ctx.print(comments)
         ctx.print(f"{name}: {translatable};")
     elif gir is None or gir.properties.get(name) is None:
-        ctx.print(f'{name}: "{escape_quote(cdata)}";')
+        ctx.print(f"{name}: {escape_quote(cdata)};")
     else:
         ctx.print_attribute(name, cdata, gir.properties.get(name).type)
     return gir
