@@ -18,16 +18,44 @@
 # SPDX-License-Identifier: LGPL-3.0-or-later
 
 
-from .. import gir
-from ..ast_utils import AstNode, validate, docs
-from ..errors import CompileError, MultipleErrors
-from ..completions_utils import *
 from .. import decompiler as decompile
-from ..decompiler import DecompileCtx, decompiler
-from ..gir import StringType, BoolType, IntType, FloatType, GirType, Enumeration
-from ..lsp_utils import Completion, CompletionItemKind, SemanticToken, SemanticTokenType
+from .. import gir
+from ..ast_utils import AstNode, context, docs, validate
+from ..completions_utils import *
+from ..decompiler import (
+    DecompileCtx,
+    decompile_translatable,
+    decompiler,
+    escape_quote,
+    truthy,
+)
+from ..errors import (
+    CodeAction,
+    CompileError,
+    CompileWarning,
+    DeprecatedWarning,
+    MultipleErrors,
+    UpgradeWarning,
+)
+from ..gir import (
+    BoolType,
+    Enumeration,
+    ExternType,
+    FloatType,
+    GirType,
+    IntType,
+    StringType,
+)
+from ..lsp_utils import (
+    Completion,
+    CompletionItemKind,
+    DocumentSymbol,
+    LocationLink,
+    SemanticToken,
+    SemanticTokenType,
+    SymbolKind,
+)
 from ..parse_tree import *
 
-
 OBJECT_CONTENT_HOOKS = AnyOf()
-VALUE_HOOKS = AnyOf()
+LITERAL = AnyOf()
